import os
import boto3
import botocore
import logging
import base64

import aws_encryption_sdk
from aws_encryption_sdk import CommitmentPolicy
from aws_encryption_sdk.exceptions import AWSEncryptionSDKClientError

from aws_cryptographic_material_providers.mpl import AwsCryptographicMaterialProviders
from aws_cryptographic_material_providers.mpl.config import MaterialProvidersConfig
from aws_cryptographic_material_providers.mpl.models import CreateAwsKmsKeyringInput
from aws_cryptographic_material_providers.mpl.references import IKeyring

from mind_castle.secret_store_base import SecretStoreBase, MindCastleSecret
from mind_castle.exceptions import (
    RetrieveSecretException,
    CreateSecretException,
)

logger = logging.getLogger(__name__)


class AWSKMSSecretStore(SecretStoreBase):
    """
    Uses AWS KMS to encrypt secrets.
    """

    store_type = "awskms"
    required_config = [
        [
            "MIND_CASTLE_AWS_REGION",
            "MIND_CASTLE_AWS_KMS_KEY_ID",
            "MIND_CASTLE_AWS_SECRET_ACCESS_KEY",
            "MIND_CASTLE_AWS_KMS_KEY_ARN",
        ],
        [
            "MIND_CASTLE_AWS_REGION",
            "MIND_CASTLE_AWS_USE_ENV_AUTH",
            "MIND_CASTLE_AWS_KMS_KEY_ARN",
        ],
    ]
    optional_config = []

    KMS_ENCRYPTION_ALGORITHM = "RSAES_OAEP_SHA_256"

    def __init__(self):
        if all([os.environ.get(req) for req in self.required_config[0]]):
            # Configure with secret key
            self.client = boto3.client(
                "kms",
                region_name=os.environ["MIND_CASTLE_AWS_REGION"],
                aws_access_key_id=os.environ["MIND_CASTLE_AWS_ACCESS_KEY_ID"],
                aws_secret_access_key=os.environ["MIND_CASTLE_AWS_SECRET_ACCESS_KEY"],
            )
        else:
            # Assume the environment is configured with the correct credentials
            self.client = boto3.client(
                "kms", region_name=os.environ["MIND_CASTLE_AWS_REGION"]
            )
        self.kms_key_arn = os.environ["MIND_CASTLE_AWS_KMS_KEY_ARN"]

        # Encryption SDK client (enforces committing algorithm suites)
        self.esdk_client = aws_encryption_sdk.EncryptionSDKClient(
            commitment_policy=CommitmentPolicy.REQUIRE_ENCRYPT_REQUIRE_DECRYPT
        )

        # MPL KMS keyring for envelope encryption
        mat_prov: AwsCryptographicMaterialProviders = AwsCryptographicMaterialProviders(
            config=MaterialProvidersConfig()
        )
        keyring_input: CreateAwsKmsKeyringInput = CreateAwsKmsKeyringInput(
            kms_key_id=self.kms_key_arn,  # KMS key identifier (ARN is valid here)
            kms_client=self.client,
        )
        self.keyring: IKeyring = mat_prov.create_aws_kms_keyring(input=keyring_input)

    def retrieve_secret(self, secret: MindCastleSecret) -> str:
        encrypted_value = secret.encrypted_value

        # For a brief period we encrypted directly. This is to support legacy secrets.
        if not secret.metadata.get("envelope_encryption"):
            try:
                response = self.client.decrypt(
                    KeyId=self.kms_key_arn,
                    EncryptionAlgorithm=self.KMS_ENCRYPTION_ALGORITHM,
                    CiphertextBlob=base64.b64decode(encrypted_value),
                )
            except botocore.exceptions.ClientError as e:
                logger.exception(f"Error retrieving secret {secret.key}:")
                raise RetrieveSecretException(e, encrypted_value)
            plaintext = response.get("Plaintext")

        else:
            try:
                plaintext, _ = self.esdk_client.decrypt(
                    source=base64.b64decode(encrypted_value),
                    keyring=self.keyring,
                )
            except (
                AWSEncryptionSDKClientError,
                botocore.exceptions.ClientError,
                Exception,
            ) as e:
                raise RetrieveSecretException(e, encrypted_value)

        return plaintext.decode("utf-8")

    def create_secret(self, value: str) -> MindCastleSecret:
        try:
            ciphertext, _ = self.esdk_client.encrypt(
                source=value.encode("utf-8"),
                keyring=self.keyring,
            )
        except (
            AWSEncryptionSDKClientError,
            botocore.exceptions.ClientError,
            Exception,
        ) as e:
            raise CreateSecretException(e)

        return MindCastleSecret(
            mind_castle_secret_type=self.store_type,
            encrypted_value=base64.b64encode(ciphertext).decode("utf-8"),
            metadata={"envelope_encryption": True},
        )

    def update_secret(self, secret: MindCastleSecret, value: str) -> MindCastleSecret:
        new_secret = self.create_secret(value)
        new_secret.key = secret.key
        return new_secret

    def delete_secret(self, secret: MindCastleSecret) -> None:
        pass



import glob
import logging
import os
import json
from typing import Any
from os.path import basename, dirname, isfile, join

from mind_castle.secret_store_base import SecretStoreBase, MindCastleSecret
from mind_castle.exceptions import RetrieveSecretException

import importlib

# Dynamically import all .py files in the current package except __init__.py
modules = glob.glob(join(dirname(__file__), "*.py"))
for f in modules:
    if isfile(f) and not f.endswith("__init__.py"):
        module_name = basename(f)[:-3]
        importlib.import_module(f".{module_name}", package=__name__)


logger = logging.getLogger(__name__)

# Get all env vars with our prefix
env_vars = {k: v for k, v in os.environ.items() if k.startswith("MIND_CASTLE_")}

stores = {}

# Populate 'stores' dict with instantiated stores
for subclass in SecretStoreBase.__subclasses__():
    configured = False
    for config_set in subclass.required_config:
        if all(e in env_vars.keys() for e in config_set):
            stores[subclass.store_type] = subclass()
            secret_store = subclass()
            logger.debug(f"Configured secret store '{subclass.store_type}'")
            configured = True

        elif not configured and any(e in env_vars.keys() for e in config_set):
            # Log a warning if some but not all required config is present
            logger.warning(
                f"Missing required config for store '{subclass.store_type}'. Requires: {config_set}"
            )


def create_secret(value: Any, secret_type: str) -> MindCastleSecret:
    value = json.dumps(value)
    secret_store = stores.get(secret_type)
    if secret_store is None:
        raise ValueError(
            f"Trying to put secret to unknown/unconfigured store '{secret_type}'"
        )

    return secret_store.create_secret(value)


def retrieve_secret(secret: MindCastleSecret) -> Any:
    secret_type = secret.mind_castle_secret_type
    if secret_type is None:
        raise ValueError("No mind_castle_secret_type provided in secret")

    secret_store = stores.get(secret_type)
    if secret_store is None:
        raise ValueError(
            f"Trying to retrieve secret from unknown/unconfigured store '{secret_type}'"
        )
    response = secret_store.retrieve_secret(secret)
    if response is None:
        raise RetrieveSecretException(
            None,
            f"Secret of type {secret_type} with key {secret.key} not found.",
        )
    return json.loads(response)


def update_secret(secret: MindCastleSecret, value: str) -> MindCastleSecret:
    secret_type = secret.mind_castle_secret_type
    if secret_type is None:
        raise ValueError("No mind_castle_secret_type provided in data")

    secret_store = stores.get(secret_type)
    if secret_store is None:
        raise ValueError(
            f"Trying to put secret to unknown/unconfigured store '{secret_type}'"
        )

    return secret_store.update_secret(secret, value)


def delete_secret(secret: MindCastleSecret) -> None:
    secret_type = secret.mind_castle_secret_type
    if secret_type is None:
        raise ValueError("No mind_castle_secret_type provided in data")

    secret_store = stores.get(secret_type)
    if secret_store is None:
        raise ValueError(
            f"Trying to retrieve secret from unknown/unconfigured store '{secret_type}'"
        )
    return secret_store.delete_secret(secret)



import logging
import wrapt
import copy

from sqlalchemy import types


from mind_castle.secret_store_base import MindCastleSecret
from mind_castle.stores import (
    create_secret,
    retrieve_secret,
)

logger = logging.getLogger(__name__)


# TODO: for remote stores
#
# @event.listens_for(Mapper, "mapper_configured")
# def _register_secret_set_event(mapper, class_):
#     """Automatically register an attribute setter function on SecretData columns"""
#     for prop in mapper.iterate_properties:
#         col = getattr(prop, "columns", [None])[0]
#         if col is not None and isinstance(col.type, SecretData):
#             # prop.key is the attribute name, e.g. "api_key" or "data"
#             attr = getattr(class_, prop.key)
#             event.listen(attr, "set", _receive_secret_set, retval=True)


class MindCastleProxy(wrapt.ObjectProxy):
    """
    Wraps any Python value + its secret details.
    """

    def __init__(self, secret_value, secret_details):
        super(MindCastleProxy, self).__init__(secret_value)
        self._self_secret_details = secret_details

    def __deepcopy__(self, memo):
        return MindCastleProxy(
            copy.deepcopy(self.__wrapped__), self._self_secret_details
        )


class SecretData(types.TypeDecorator):
    impl = types.JSON
    cache_ok = True

    def __init__(self, store_type: str, secret_key_prefix: str = "", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.secret_type = store_type
        self.secret_key_prefix = secret_key_prefix

    # Mutate/encrypt a value for storage in the DB
    def process_bind_param(self, value, dialect):
        if value is None:
            # Don't waste our time encrypting None
            return None
        if self.secret_type == "none":
            return value
        if isinstance(value, MindCastleProxy):
            # This value is a secret previously stored in the DB.
            # We can just return the secret details already calculated
            return value._self_secret_details.model_dump(mode="json")

        # Default: create a new secret from the value
        return create_secret(value, self.secret_type).model_dump(mode="json")

    # Decrypt a value after retrieval from the DB
    def process_result_value(self, value, dialect):
        if value is None:
            return None

        if not isinstance(value, dict) or value.get("mind_castle_secret_type") is None:
            # This isn't a mind castle secret, so treat it like a raw value
            logger.debug("Got a plaintext value from the DB - returning it as-is")
            return value

        # Default: load the json data into a MindCastleSecret, then retrieve the plaintext
        secret = MindCastleSecret(**value)
        secret_plaintext = retrieve_secret(secret)
        return MindCastleProxy(secret_plaintext, secret)


# def _receive_secret_set(target, value, oldvalue, initiator):
#     """Handle setting a new SecretData value, while maintaining existing secret info"""
#     # if the thing you’re overwriting was already a proxy,
#     # just swap out its wrapped value, keep its secret ID, and return it.

#     logger.debug(f"RECEIVE SECRET SET: {oldvalue} -> {value}")
#     logger.debug(f"{target} | {initiator}")
#     if isinstance(oldvalue, MindCastleProxy) and oldvalue != value:
#         logger.debug(f"SETTING EXISTING SECRET: {oldvalue} -> {value}")
#         oldvalue.__wrapped__ = value  # change the “plaintext”
#         flag_modified(target, initiator.key)
#         return oldvalue  # SQLAlchemy will leave the attribute as-is
#     # otherwise it’s the first time setting, so just hand back the raw value;
#     # your before_flush hook will pick it up and wrap it later.
#     return value


# @event.listens_for(Session, "before_flush")
# def _track_secret_changes(session: Session, flush_context, instances):
#     # TODO: If an error occurs here with any secret interactions, explain implications to the user
#     logger.debug("================================= BEFORE FLUSH")
#     info = session.info
#     info.setdefault("pending_new", [])
#     info.setdefault("pending_old", [])
#     info.setdefault("pending_deleted", [])

#     # Handle INSERTs and UPDATEs
#     logger.debug(f"New or dirty objects: {list(session.new) + list(session.dirty)}")
#     for obj in list(session.new) + list(session.dirty):
#         state = inspect(obj)
#         is_update = obj in session.dirty
#         logger.debug(f"is_update: {is_update}")
#         for attr in state.mapper.column_attrs:
#             logger.debug(f"attr: {attr}")
#             col = attr.columns[0]
#             if not isinstance(col.type, SecretData):
#                 logger.debug("- is not secret data")
#                 continue
#             logger.debug("- is secret data")
#             secret_type = col.type.secret_type
#             logger.debug(f"- is a {secret_type} secret")
#             hist = state.get_history(attr.key, True)
#             if not hist.has_changes():
#                 logger.debug("- does not have changes")
#                 continue

#             new_value = hist.added[0]
#             logger.debug(f"- has new value: {new_value}")
#             if is_update and isinstance(new_value, MindCastleProxy):
#                 logger.debug("- has been updated")
#                 # Ideally we would call update_secret like below:
#                 # UPDATE in place: same ID, just update payload
#                 # update_secret(new_value._self_secret_details, new_value.__wrapped__)
#                 # sid = new_value._self_secret_details

#                 # For now we're going to create a new secret here and mark the old one for deletion.
#                 # This is because we can't guarantee an update will go through if we wait till after DB commit
#                 # and we can't guarantee the DB commit will go through if we update the secret here.

#                 # Maybe in future we will support updates/secret history, but for now we create a secret with the new value here
#                 # and delete the old secret after DB commit
#                 # This is a bit out of whack with how we handle updates to sqlalchemy attrs, as there we update the secret proxy value in-place
#                 sid = create_secret(
#                     new_value.__wrapped__,
#                     new_value._self_secret_details["mind_castle_secret_type"],
#                 )
#                 info["pending_new"].append(sid)
#                 info["pending_old"].append(new_value._self_secret_details)
#             else:
#                 logger.debug("- has been created")
#                 # brand‐new secret
#                 sid = create_secret(new_value, secret_type)
#                 logger.debug(f"- created secret: {sid}")
#                 info["pending_new"].append(sid)
#                 if hist.deleted and isinstance(hist.deleted[0], MindCastleProxy):
#                     info["pending_old"].append(hist.deleted[0]._self_secret_details)

#             # wrap so that process_bind_param emits sid
#             setattr(obj, attr.key, MindCastleProxy(new_value, sid))

#     # Handle DELETEs
#     for obj in list(session.deleted):
#         state = inspect(obj)
#         for attr in state.mapper.column_attrs:
#             col = attr.columns[0]
#             if isinstance(col.type, SecretData):
#                 val = getattr(obj, attr.key)
#                 if isinstance(val, MindCastleProxy):
#                     info["pending_deleted"].append(val._self_secret_details)


# @event.listens_for(Session, "after_commit")
# def _on_commit_cleanup(session: Session):
#     logger.debug("================================= AFTER COMMIT")
#     # Only after a successful commit do we delete old/removed secrets
#     # TODO: If an error occurs here, explain implications to the user
#     for sid in session.info.get("pending_old", []):
#         delete_secret(sid)
#     for sid in session.info.get("pending_deleted", []):
#         delete_secret(sid)
#     session.info.pop("pending_new", None)
#     session.info.pop("pending_old", None)
#     session.info.pop("pending_deleted", None)


# @event.listens_for(Session, "after_rollback")
# def _on_rollback_cleanup(session: Session):
#     logger.debug("================================= AFTER rollback")
#     # If we roll back, remove any NEW secrets that never made it into the DB
#     # TODO: If an error occurs here, explain implications to the user
#     for sid in session.info.get("pending_new", []):
#         delete_secret(sid)
#     session.info.pop("pending_new", None)
#     session.info.pop("pending_old", None)
#     session.info.pop("pending_deleted", None)



from sqlalchemy import create_engine
from sqlalchemy.orm import Session
import pytest
import logging
from moto import mock_aws
import os
import base64
import math

from hypothesis import given
from hypothesis_jsonschema import from_schema

os.environ["MIND_CASTLE_LOCAL_ENCRYPTION_SECRET"] = base64.urlsafe_b64encode(
    os.urandom(16)
).decode()

from mind_castle.stores import retrieve_secret
from mind_castle.exceptions import RetrieveSecretException
from models import (
    Base,
    SimpleNoneModel,
    SimpleLocalEncryptionModel,
)

logger = logging.getLogger(__name__)

# TEST_STORES = [SimpleMemoryModel, SimpleAWSSecretsManagerModel, SimpleAWSKMSModel, SimpleJsonModel, SimpleLocalEncryptionModel]
TEST_STORES = [SimpleNoneModel, SimpleLocalEncryptionModel]


# Create an in-memory SQLite database
engine = create_engine("sqlite://", echo=True)
Base.metadata.create_all(engine)

session = Session(engine)

mock = mock_aws()
mock.start()


@pytest.mark.parametrize("model", TEST_STORES)
@given(from_schema({}, allow_x00=False, codec="utf-8"))
def test_create_retrieve_secret_values(model, secret_data):
    test_object = model(data=secret_data)
    session.add(test_object)
    session.commit()
    session.expire_all()  # Get rid of cache

    item = session.get(model, test_object.id)
    if isinstance(secret_data, float) or isinstance(secret_data, int):
        assert math.isclose(item.data, secret_data)
    else:
        assert item.data == secret_data


@pytest.mark.parametrize("model", TEST_STORES)
@given(
    from_schema({}, allow_x00=False, codec="utf-8"),
    from_schema({}, allow_x00=False, codec="utf-8"),
)
def test_update_secret_values(model, secret_data, updated_data):
    test_object = model(data=secret_data)
    session.add(test_object)
    session.commit()
    session.expire_all()  # Get rid of cache

    # Update the value
    test_object.data = updated_data
    session.commit()
    session.expire_all()  # Get rid of cache

    test_object = session.get(model, test_object.id)
    if isinstance(updated_data, float) or isinstance(updated_data, int):
        assert math.isclose(test_object.data, updated_data)
    else:
        assert test_object.data == updated_data
    # Decided to create new secrets on update for now, see comments in _track_secret_changes()
    # assert test_object.data._self_secret_details["key"] == key_before


@pytest.mark.parametrize("model", TEST_STORES)
@given(from_schema({}, allow_x00=False, codec="utf-8"))
def test_delete_secret_values(model, secret_data):
    test_object = model(data=secret_data)
    session.add(test_object)
    session.commit()
    session.expire_all()  # Get rid of cache

    # Delete the object
    session.delete(test_object)
    session.commit()
    session.expire_all()  # Get rid of cache

    assert not session.get(model, test_object.id)

    # None is a valid value that's output to shortcut encrypting nothing
    # Local storage store types (KMS etc) wont throw an error here
    if test_object.data and model not in [SimpleLocalEncryptionModel, SimpleNoneModel]:
        with pytest.raises(RetrieveSecretException):
            retrieve_secret(test_object.data._self_secret_details)
